@@ -1,6 +1,8 @@
 """Tests related to uaclient.entitlement.base module."""
 
+import contextlib
 import copy
+import io
 import logging
 import mock
 from functools import partial
@@ -23,7 +25,8 @@ from uaclient.entitlements.livepatch import (
 )
 from uaclient.entitlements.tests.conftest import machine_token
 from uaclient import status
-from uaclient.status import ContractStatus
+from uaclient.status import ApplicationStatus, ContractStatus
+from uaclient.util import ProcessExecutionError
 
 PLATFORM_INFO_SUPPORTED = MappingProxyType(
     {
@@ -35,6 +38,9 @@ PLATFORM_INFO_SUPPORTED = MappingProxyType(
 )
 
 M_PATH = "uaclient.entitlements.livepatch."  # mock path
+M_LIVEPATCH_STATUS = M_PATH + "LivepatchEntitlement.application_status"
+DISABLED_APP_STATUS = (status.ApplicationStatus.DISABLED, "")
+
 M_BASE_PATH = "uaclient.entitlements.base.UAEntitlement."
 
 DEFAULT_AFFORDANCES = {
@@ -172,6 +178,7 @@ class TestLivepatchProcessConfigDirectives:
         assert 0 == m_subp.call_count
 
 
+@mock.patch(M_LIVEPATCH_STATUS, return_value=DISABLED_APP_STATUS)
 @mock.patch(
     "uaclient.entitlements.livepatch.util.is_container", return_value=False
 )
@@ -181,7 +188,12 @@ class TestLivepatchEntitlementCanEnable:
         ("4.4.0-00-generic", "5.0.0-00-generic", "4.19.0-00-generic"),
     )
     def test_can_enable_true_on_entitlement_inactive(
-        self, _m_is_container, supported_kernel_ver, capsys, entitlement
+        self,
+        _m_is_container,
+        _m_livepatch_status,
+        supported_kernel_ver,
+        capsys,
+        entitlement,
     ):
         """When entitlement is INACTIVE, can_enable returns True."""
         supported_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -195,7 +207,7 @@ class TestLivepatchEntitlementCanEnable:
         assert [mock.call()] == m_container.call_args_list
 
     def test_can_enable_false_on_unsupported_kernel_min_version(
-        self, _m_is_container, capsys, entitlement
+        self, _m_is_container, _m_livepatch_status, capsys, entitlement
     ):
         """"False when on a kernel less or equal to minKernelVersion."""
         unsupported_min_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -211,7 +223,7 @@ class TestLivepatchEntitlementCanEnable:
         assert (msg, "") == capsys.readouterr()
 
     def test_can_enable_false_on_unsupported_kernel_flavor(
-        self, _m_is_container, capsys, entitlement
+        self, _m_is_container, _m_livepatch_status, capsys, entitlement
     ):
         """"When on an unsupported kernel, can_enable returns False."""
         unsupported_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -239,6 +251,7 @@ class TestLivepatchEntitlementCanEnable:
     def test_can_enable_false_on_unsupported_min_kernel_version(
         self,
         _m_is_container,
+        _m_livepatch_status,
         kernel_version,
         meets_min_version,
         capsys,
@@ -264,7 +277,7 @@ class TestLivepatchEntitlementCanEnable:
         assert (msg, "") == capsys.readouterr()
 
     def test_can_enable_false_on_unsupported_architecture(
-        self, _m_is_container, capsys, entitlement
+        self, _m_is_container, _m_livepatch_status, capsys, entitlement
     ):
         """"When on an unsupported architecture, can_enable returns False."""
         unsupported_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -279,7 +292,7 @@ class TestLivepatchEntitlementCanEnable:
         assert (msg, "") == capsys.readouterr()
 
     def test_can_enable_false_on_containers(
-        self, m_is_container, capsys, entitlement
+        self, m_is_container, _m_livepatch_status, capsys, entitlement
     ):
         """When is_container is True, can_enable returns False."""
         unsupported_min_kernel = copy.deepcopy(dict(PLATFORM_INFO_SUPPORTED))
@@ -404,10 +417,12 @@ class TestLivepatchEntitlementEnable:
             ["apt-get", "install", "--assume-yes", "snapd"],
             capture=True,
             retry_sleeps=apt.APT_RETRIES,
-        ),
+        )
+    ]
+    mocks_snap_wait_seed = [
         mock.call(
             ["/usr/bin/snap", "wait", "system", "seed.loaded"], capture=True
-        ),
+        )
     ]
     mocks_livepatch_install = [
         mock.call(
@@ -416,7 +431,9 @@ class TestLivepatchEntitlementEnable:
             retry_sleeps=[0.5, 1, 5],
         )
     ]
-    mocks_install = mocks_snapd_install + mocks_livepatch_install
+    mocks_install = (
+        mocks_snapd_install + mocks_snap_wait_seed + mocks_livepatch_install
+    )
     mocks_config = [
         mock.call(
             [
@@ -535,7 +552,9 @@ class TestLivepatchEntitlementEnable:
         m_app_status.return_value = application_status, "enabled"
         assert entitlement.enable()
         assert (
-            self.mocks_livepatch_install + self.mocks_config
+            self.mocks_snap_wait_seed
+            + self.mocks_livepatch_install
+            + self.mocks_config
             in m_subp.call_args_list
         )
         msg = (
@@ -629,3 +648,80 @@ class TestLivepatchEntitlementEnable:
         ]
         assert subp_no_livepatch_disable == m_subp.call_args_list
         assert ("Canonical livepatch enabled.\n", "") == capsys.readouterr()
+
+    @pytest.mark.parametrize(
+        "cls_name, cls_title",
+        (
+            ("FIPSEntitlement", "FIPS"),
+            ("FIPSUpdatesEntitlement", "FIPS Updates"),
+        ),
+    )
+    @mock.patch("uaclient.entitlements.repo.handle_message_operations")
+    @mock.patch("uaclient.util.is_container", return_value=False)
+    def test_enable_fails_when_blocking_service_is_enabled(
+        self,
+        m_is_container,
+        m_handle_message_op,
+        cls_name,
+        cls_title,
+        entitlement,
+    ):
+        m_handle_message_op.return_value = True
+
+        fake_stdout = io.StringIO()
+        with mock.patch(M_LIVEPATCH_STATUS, return_value=DISABLED_APP_STATUS):
+            with mock.patch(
+                "uaclient.entitlements.fips.{}.application_status".format(
+                    cls_name
+                )
+            ) as m_fips:
+                m_fips.return_value = (status.ApplicationStatus.ENABLED, "")
+                with contextlib.redirect_stdout(fake_stdout):
+                    entitlement.enable()
+
+        expected_msg = "Cannot enable Livepatch when {} is enabled".format(
+            cls_title
+        )
+        assert expected_msg.strip() == fake_stdout.getvalue().strip()
+
+
+class TestLivepatchApplicationStatus:
+    @pytest.mark.parametrize("which_result", ((True), (False)))
+    @pytest.mark.parametrize("subp_raise_exception", ((True), (False)))
+    @mock.patch("uaclient.util.which")
+    @mock.patch("uaclient.util.subp")
+    def test_application_status(
+        self, m_subp, m_which, subp_raise_exception, which_result, entitlement
+    ):
+        m_which.return_value = which_result
+
+        if subp_raise_exception:
+            m_subp.side_effect = ProcessExecutionError("error msg")
+
+        status, details = entitlement.application_status()
+
+        if not which_result:
+            assert status == ApplicationStatus.DISABLED
+            assert "canonical-livepatch snap is not installed." in details
+        elif subp_raise_exception:
+            assert status == ApplicationStatus.DISABLED
+            assert "error msg" in details
+        else:
+            assert status == ApplicationStatus.ENABLED
+            assert "" == details
+
+    @mock.patch("time.sleep")
+    @mock.patch("uaclient.util.which", return_value=True)
+    def test_status_command_retry_on_application_status(
+        self, m_which, m_sleep, entitlement
+    ):
+        from uaclient import util
+
+        with mock.patch.object(util, "_subp") as m_subp:
+            m_subp.side_effect = ProcessExecutionError("error msg")
+            status, details = entitlement.application_status()
+
+            assert m_subp.call_count == 3
+            assert m_sleep.call_count == 2
+            assert status == ApplicationStatus.DISABLED
+            assert "error msg" in details
