@@ -1,5 +1,4 @@
 import copy
-from textwrap import dedent
 
 import mock
 import pytest
@@ -7,7 +6,10 @@ from types import MappingProxyType
 
 from uaclient import apt
 from uaclient import config
-from uaclient.entitlements.repo import RepoEntitlement
+from uaclient.entitlements.repo import (
+    RepoEntitlement,
+    handle_message_operations,
+)
 from uaclient.entitlements.tests.conftest import machine_token
 from uaclient import exceptions
 from uaclient import status
@@ -15,6 +17,7 @@ from uaclient import util
 
 
 M_PATH = "uaclient.entitlements.repo."
+M_CONTRACT_PATH = "uaclient.entitlements.repo.contract.UAContractClient."
 
 PLATFORM_INFO_SUPPORTED = MappingProxyType(
     {
@@ -222,51 +225,79 @@ class TestProcessContractDeltas:
         )
         assert expected_msg in caplog_text()
 
+    @pytest.mark.parametrize("packages", ([], ["extremetuxracer"]))
+    @mock.patch.object(RepoTestEntitlement, "install_packages")
     @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
     @mock.patch.object(RepoTestEntitlement, "setup_apt_config")
     @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
     @mock.patch.object(RepoTestEntitlement, "application_status")
-    def test_update_apt_config_when_active(
+    @mock.patch.object(RepoTestEntitlement, "_check_apt_url_is_applied")
+    def test_update_apt_config_and_install_packages_when_active(
         self,
+        m_check_apt_url_applied,
         m_application_status,
         m_remove_apt_config,
         m_setup_apt_config,
         m_remove_auth_apt_repo,
+        m_install_packages,
+        packages,
         entitlement,
     ):
-        """Update_apt_config when service is active and not enableByDefault."""
+        """Update_apt_config and packages if active and not enableByDefault."""
+        m_check_apt_url_applied.return_value = False
         application_status = status.ApplicationStatus.ENABLED
         m_application_status.return_value = (application_status, "")
+        deltas = {
+            "entitlement": {"obligations": {"enableByDefault": False}},
+            "resourceToken": "repotest-token",
+        }
+        if packages:
+            deltas["entitlement"] = {
+                "directives": {"additionalPackages": packages}
+            }
+
         assert entitlement.process_contract_deltas(
-            {"entitlement": {"entitled": True}},
-            {
-                "entitlement": {"obligations": {"enableByDefault": False}},
-                "resourceToken": "repotest-token",
-            },
+            {"entitlement": {"entitled": True}}, deltas
         )
         assert [mock.call()] == m_remove_apt_config.call_args_list
         assert [mock.call()] == m_setup_apt_config.call_args_list
+        if packages:
+            assert [
+                mock.call(package_list=packages)
+            ] == m_install_packages.call_args_list
+        else:
+            assert 0 == m_install_packages.call_count
         assert [] == m_remove_auth_apt_repo.call_args_list
+        assert 1 == m_check_apt_url_applied.call_count
 
+    @mock.patch(
+        "uaclient.entitlements.base.UAEntitlement.process_contract_deltas"
+    )
+    @mock.patch("uaclient.config.UAConfig.read_cache")
     @mock.patch(
         M_PATH + "util.get_platform_info", return_value={"series": "trusty"}
     )
     @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
     @mock.patch.object(RepoTestEntitlement, "setup_apt_config")
     @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
-    @mock.patch.object(RepoTestEntitlement, "application_status")
+    @mock.patch.object(RepoTestEntitlement, "_check_apt_url_is_applied")
     def test_remove_old_auth_apt_repo_when_active_and_apt_url_delta(
         self,
-        m_application_status,
+        m_check_apt_url_applied,
         m_remove_apt_config,
         m_setup_apt_config,
         m_remove_auth_apt_repo,
         m_platform_info,
+        m_read_cache,
+        m_process_contract_deltas,
         entitlement,
     ):
         """Remove old apt url when aptURL delta occurs on active service."""
-        application_status = status.ApplicationStatus.ENABLED
-        m_application_status.return_value = (application_status, "")
+        m_check_apt_url_applied.return_value = False
+        m_process_contract_deltas.return_value = False
+        m_read_cache.return_value = {
+            "services": [{"name": "repotest", "status": "enabled"}]
+        }
         assert entitlement.process_contract_deltas(
             {
                 "entitlement": {
@@ -286,18 +317,78 @@ class TestProcessContractDeltas:
         assert [mock.call()] == m_setup_apt_config.call_args_list
         apt_auth_remove_calls = [
             mock.call(
-                "/etc/apt/sources.list.d/ubuntu-repotest-trusty.list",
-                "http://old",
-            )
-        ]
-        assert apt_auth_remove_calls == m_remove_auth_apt_repo.call_args_list
-        apt_auth_remove_calls = [
-            mock.call(
-                "/etc/apt/sources.list.d/ubuntu-repotest-trusty.list",
-                "http://old",
+                "/etc/apt/sources.list.d/ubuntu-repotest.list", "http://old"
             )
         ]
         assert apt_auth_remove_calls == m_remove_auth_apt_repo.call_args_list
+        assert [
+            mock.call("status-cache"),
+            mock.call("status-cache"),
+        ] == m_read_cache.call_args_list
+        assert 1 == m_process_contract_deltas.call_count
+
+        assert [
+            mock.call("http://new")
+        ] == m_check_apt_url_applied.call_args_list
+        assert 1 == m_check_apt_url_applied.call_count
+
+    @mock.patch(
+        "uaclient.entitlements.base.UAEntitlement.process_contract_deltas"
+    )
+    @mock.patch("uaclient.config.UAConfig.read_cache")
+    @mock.patch(
+        M_PATH + "util.get_platform_info", return_value={"series": "trusty"}
+    )
+    @mock.patch(M_PATH + "apt.remove_auth_apt_repo")
+    @mock.patch.object(RepoTestEntitlement, "setup_apt_config")
+    @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
+    @mock.patch.object(RepoTestEntitlement, "_check_apt_url_is_applied")
+    def test_system_does_not_change_when_apt_url_delta_already_applied(
+        self,
+        m_check_apt_url_applied,
+        m_remove_apt_config,
+        m_setup_apt_config,
+        m_remove_auth_apt_repo,
+        m_platform_info,
+        m_read_cache,
+        m_process_contract_deltas,
+        entitlement,
+    ):
+        """Do not change system if apt url delta is already applied."""
+        m_check_apt_url_applied.return_value = True
+        m_process_contract_deltas.return_value = False
+        m_read_cache.return_value = {
+            "services": [{"name": "repotest", "status": "enabled"}]
+        }
+        assert entitlement.process_contract_deltas(
+            {
+                "entitlement": {
+                    "entitled": True,
+                    "directives": {"aptURL": "http://old"},
+                }
+            },
+            {
+                "entitlement": {
+                    "obligations": {"enableByDefault": False},
+                    "directives": {"aptURL": "http://new"},
+                },
+                "resourceToken": "repotest-token",
+            },
+        )
+        assert m_remove_apt_config.call_count == 0
+        assert m_setup_apt_config.call_count == 0
+        assert m_remove_auth_apt_repo.call_count == 0
+
+        assert [
+            mock.call("status-cache"),
+            mock.call("status-cache"),
+        ] == m_read_cache.call_args_list
+        assert 1 == m_process_contract_deltas.call_count
+
+        assert [
+            mock.call("http://new")
+        ] == m_check_apt_url_applied.call_args_list
+        assert 1 == m_check_apt_url_applied.call_count
 
 
 class TestRepoEnable:
@@ -318,8 +409,90 @@ class TestRepoEnable:
         expected_call = mock.call(silent=bool(silent_if_inapplicable))
         assert [expected_call] == m_can_enable.call_args_list
 
+    @pytest.mark.parametrize(
+        "pre_enable_msg, output, can_enable_call_count",
+        (
+            (["msg1", (lambda: False, {}), "msg2"], "msg1\n", 0),
+            (["msg1", (lambda: True, {}), "msg2"], "msg1\nmsg2\n", 1),
+        ),
+    )
+    @mock.patch.object(RepoTestEntitlement, "can_enable", return_value=False)
+    def test_enable_can_exit_on_pre_enable_messaging_hooks(
+        self,
+        m_can_enable,
+        pre_enable_msg,
+        output,
+        can_enable_call_count,
+        entitlement,
+        capsys,
+    ):
+        with mock.patch(
+            M_PATH + "RepoEntitlement.messaging",
+            new_callable=mock.PropertyMock,
+        ) as m_messaging:
+            m_messaging.return_value = {"pre_enable": pre_enable_msg}
+            with mock.patch.object(type(entitlement), "packages", []):
+                entitlement.enable()
+        stdout, _ = capsys.readouterr()
+        assert output == stdout
+        assert can_enable_call_count == m_can_enable.call_count
+
+    @pytest.mark.parametrize(
+        "pre_disable_msg,post_disable_msg,output,retval",
+        (
+            (
+                ["pre1", (lambda: False, {}), "pre2"],
+                ["post1"],
+                "pre1\n",
+                False,
+            ),
+            (["pre1", (lambda: True, {}), "pre2"], [], "pre1\npre2\n", True),
+            (
+                ["pre1", (lambda: True, {}), "pre2"],
+                ["post1", (lambda: False, {}), "post2"],
+                "pre1\npre2\npost1\n",
+                False,
+            ),
+            (
+                ["pre1", (lambda: True, {}), "pre2"],
+                ["post1", (lambda: True, {}), "post2"],
+                "pre1\npre2\npost1\npost2\n",
+                True,
+            ),
+        ),
+    )
+    @mock.patch(M_PATH + "util.subp", return_value=("", ""))
+    @mock.patch(M_PATH + "util.should_reboot")
+    @mock.patch.object(RepoTestEntitlement, "remove_apt_config")
+    @mock.patch.object(RepoTestEntitlement, "can_disable", return_value=True)
+    def test_enable_can_exit_on_pre_or_post_disable_messaging_hooks(
+        self,
+        _can_disable,
+        remove_apt_config,
+        m_should_reboot,
+        m_subp,
+        pre_disable_msg,
+        post_disable_msg,
+        output,
+        retval,
+        entitlement,
+        capsys,
+    ):
+        messaging = {
+            "pre_disable": pre_disable_msg,
+            "post_disable": post_disable_msg,
+        }
+        m_should_reboot.return_value = False
+        with mock.patch.object(type(entitlement), "messaging", messaging):
+            with mock.patch.object(type(entitlement), "packages", []):
+                assert retval is entitlement.disable()
+        stdout, _ = capsys.readouterr()
+        assert output == stdout
+
+    @pytest.mark.parametrize("should_reboot", (False, True))
     @pytest.mark.parametrize("with_pre_install_msg", (False, True))
     @pytest.mark.parametrize("packages", (["a"], [], None))
+    @mock.patch(M_PATH + "util.should_reboot")
     @mock.patch(M_PATH + "util.subp", return_value=("", ""))
     @mock.patch(M_PATH + "apt.add_auth_apt_repo")
     @mock.patch(M_PATH + "os.path.exists", return_value=True)
@@ -332,20 +505,25 @@ class TestRepoEnable:
         m_exists,
         m_apt_add,
         m_subp,
+        m_should_reboot,
         entitlement,
         capsys,
         caplog_text,
         tmpdir,
         packages,
         with_pre_install_msg,
+        should_reboot,
     ):
         """On enable add authenticated apt repo and refresh package lists."""
         m_platform.return_value = {"series": "xenial"}
+        m_should_reboot.return_value = should_reboot
 
         pre_install_msgs = ["Some pre-install information", "Some more info"]
         if with_pre_install_msg:
-            messaging_patch = mock.patch.object(
-                entitlement, "messaging", {"pre_install": pre_install_msgs}
+            messaging_patch = mock.patch(
+                M_PATH + "RepoEntitlement.messaging",
+                new_callable=mock.PropertyMock,
+                return_value={"pre_install": pre_install_msgs},
             )
         else:
             messaging_patch = mock.MagicMock()
@@ -355,14 +533,19 @@ class TestRepoEnable:
                 ["apt-get", "update"],
                 capture=True,
                 retry_sleeps=apt.APT_RETRIES,
+                env={},
             )
         ]
-        expected_output = dedent(
-            """\
-        Updating package lists
-        Repo Test Class enabled
-        """
+
+        reboot_msg = "A reboot is required to complete install"
+        expected_output = (
+            "\n".join(
+                ["Updating package lists", "Repo Test Class enabled"]
+                + ([reboot_msg] if should_reboot else [])
+            )
+            + "\n"
         )
+
         if packages is not None:
             if len(packages) > 0:
                 expected_apt_calls.append(
@@ -375,16 +558,18 @@ class TestRepoEnable:
                         ],
                         capture=True,
                         retry_sleeps=apt.APT_RETRIES,
+                        env={},
                     )
                 )
                 expected_output = (
                     "\n".join(
-                        [
-                            "Updating package lists",
-                            "Installing Repo Test Class packages",
-                        ]
+                        ["Updating package lists"]
                         + (pre_install_msgs if with_pre_install_msg else [])
-                        + ["Repo Test Class enabled"]
+                        + [
+                            "Installing Repo Test Class packages",
+                            "Repo Test Class enabled",
+                        ]
+                        + ([reboot_msg] if should_reboot else [])
                     )
                     + "\n"
                 )
@@ -404,7 +589,7 @@ class TestRepoEnable:
         assert expected_apt_calls == m_subp.call_args_list
         add_apt_calls = [
             mock.call(
-                "/etc/apt/sources.list.d/ubuntu-repotest-xenial.list",
+                "/etc/apt/sources.list.d/ubuntu-repotest.list",
                 "http://REPOTEST",
                 "repotest-token",
                 ["xenial"],
@@ -412,6 +597,7 @@ class TestRepoEnable:
             )
         ]
         assert add_apt_calls == m_apt_add.call_args_list
+        assert 1 == m_should_reboot.call_count
         stdout, _ = capsys.readouterr()
         assert expected_output == stdout
 
@@ -440,10 +626,6 @@ class TestRepoEnable:
                             entitlement.enable()
 
         assert "Could not enable Repo Test Class." == excinfo.value.msg
-        expected_call = mock.call(
-            ["apt-get", "remove", "--assume-yes"] + packages
-        )
-        assert expected_call in m_subp.call_args_list
         assert 1 == m_rac.call_count
 
 
@@ -494,7 +676,7 @@ class TestRemoveAptConfig:
         ] == m_remove_apt_list_files.call_args_list
         assert [
             mock.call(
-                "/etc/apt/sources.list.d/ubuntu-repotest-xenial.list",
+                "/etc/apt/sources.list.d/ubuntu-repotest.list",
                 "http://REPOTEST",
                 "test.gpg",
             )
@@ -527,7 +709,7 @@ class TestRemoveAptConfig:
             mock.call("http://REPOTEST")
         ] == m_remove_repo_from_apt_auth_file.call_args_list
         assert [
-            mock.call("/etc/apt/sources.list.d/ubuntu-repotest-xenial.list")
+            mock.call("/etc/apt/sources.list.d/ubuntu-repotest.list")
         ] == m_restore_commented_apt_list_file.call_args_list
 
     @mock.patch(M_PATH + "apt.add_ppa_pinning")
@@ -553,7 +735,7 @@ class TestRemoveAptConfig:
         entitlement.remove_apt_config()
         assert [
             mock.call(
-                "/etc/apt/preferences.d/ubuntu-repotest-xenial",
+                "/etc/apt/preferences.d/ubuntu-repotest",
                 "http://REPOTEST",
                 None,
                 "never",
@@ -585,9 +767,7 @@ class TestRemoveAptConfig:
         with mock.patch(M_PATH + "os.path.exists") as m_exists:
             m_exists.return_value = True
             entitlement.remove_apt_config()
-        expected_call = [
-            mock.call("/etc/apt/preferences.d/ubuntu-repotest-xenial")
-        ]
+        expected_call = [mock.call("/etc/apt/preferences.d/ubuntu-repotest")]
         assert expected_call == m_exists.call_args_list
         assert expected_call == m_unlink.call_args_list
 
@@ -628,6 +808,43 @@ class TestSetupAptConfig:
             entitlement.setup_apt_config()
         assert err_msg in str(excinfo.value)
 
+    @pytest.mark.parametrize("enable_by_default", (True, False))
+    @mock.patch(M_CONTRACT_PATH + "request_resource_machine_access")
+    @mock.patch(M_PATH + "apt.add_auth_apt_repo")
+    @mock.patch(M_PATH + "apt.run_apt_command")
+    def test_setup_apt_config_request_machine_access_when_no_resource_token(
+        self,
+        run_apt_command,
+        add_auth_apt_repo,
+        request_resource_machine_access,
+        enable_by_default,
+        entitlement_factory,
+        caplog_text,
+    ):
+        """request_machine_access routes when contract lacks resourceToken."""
+        entitlement = entitlement_factory(
+            RepoTestEntitlement,
+            affordances={"series": ["xenial"]},
+            obligations={"enableByDefault": enable_by_default},
+        )
+        machine_token = entitlement.cfg.read_cache("machine-token")
+        # Drop resourceTokens values from base machine-token.
+        machine_token["resourceTokens"] = []
+        entitlement.cfg.write_cache("machine-token", machine_token)
+        entitlement.setup_apt_config()
+        expected_msg = (
+            "No resourceToken present in contract for service Repo Test"
+            " Class. Using machine token as credentials"
+        )
+        if enable_by_default:
+            assert expected_msg in caplog_text()
+            assert 0 == request_resource_machine_access.call_count
+        else:
+            assert expected_msg not in caplog_text()
+            assert [
+                mock.call("blah", "repotest")
+            ] == request_resource_machine_access.call_args_list
+
     @mock.patch(M_PATH + "apt.add_auth_apt_repo")
     @mock.patch(M_PATH + "apt.run_apt_command")
     def test_install_prerequisite_packages(
@@ -695,12 +912,12 @@ class TestSetupAptConfig:
             m_exists.return_value = True
             entitlement.setup_apt_config()
         assert [
-            mock.call("/etc/apt/preferences.d/ubuntu-repotest-xenial"),
+            mock.call("/etc/apt/preferences.d/ubuntu-repotest"),
             mock.call("/usr/lib/apt/methods/https"),
             mock.call("/usr/sbin/update-ca-certificates"),
         ] == m_exists.call_args_list
         assert [
-            mock.call("/etc/apt/preferences.d/ubuntu-repotest-xenial")
+            mock.call("/etc/apt/preferences.d/ubuntu-repotest")
         ] == m_unlink.call_args_list
 
     @mock.patch(M_PATH + "apt.add_auth_apt_repo")
@@ -730,7 +947,7 @@ class TestSetupAptConfig:
         ] == m_exists.call_args_list
         assert [
             mock.call(
-                "/etc/apt/preferences.d/ubuntu-repotest-xenial",
+                "/etc/apt/preferences.d/ubuntu-repotest",
                 "http://REPOTEST",
                 "RepoTestOrigin",
                 entitlement.repo_pin_priority,
@@ -792,3 +1009,34 @@ class TestApplicationStatus:
             expected_explanation = "Repo Test Class is not configured"
         assert expected_status == application_status
         assert expected_explanation == explanation
+
+
+def success_call():
+    print("success")
+    return True
+
+
+def fail_call(a=None):
+    print("fail %s" % a)
+    return False
+
+
+class TestHandleMessageOperations:
+    @pytest.mark.parametrize(
+        "msg_ops, retval, output",
+        (
+            ([], True, ""),
+            (["msg1", "msg2"], True, "msg1\nmsg2\n"),
+            (
+                [(success_call, {}), "msg1", (fail_call, {"a": 1}), "msg2"],
+                False,
+                "success\nmsg1\nfail 1\n",
+            ),
+        ),
+    )
+    def test_handle_message_operations_for_strings_and_callables(
+        self, msg_ops, retval, output, capsys
+    ):
+        assert retval is handle_message_operations(msg_ops)
+        out, _err = capsys.readouterr()
+        assert output == out
