@@ -1,4 +1,6 @@
+import contextlib
 import io
+import json
 import logging
 import mock
 import os
@@ -10,17 +12,20 @@ import textwrap
 import pytest
 
 from uaclient.cli import (
+    action_help,
     assert_attached,
+    assert_lock_file,
     assert_not_attached,
     assert_root,
     get_parser,
     main,
-    require_valid_entitlement_name,
+    get_valid_entitlement_names,
     setup_logging,
 )
 
 from uaclient.exceptions import (
     AlreadyAttachedError,
+    LockHeldError,
     NonRootUserError,
     UserFacingError,
     UnattachedError,
@@ -33,23 +38,42 @@ BIG_DESC = "123456789 " * 7 + "next line"
 BIG_URL = "http://" + "adsf" * 10
 
 
-SERVICES_WRAPPED_HELP = textwrap.dedent(
+ALL_SERVICES_WRAPPED_HELP = textwrap.dedent(
     """
 Client to manage Ubuntu Advantage services on a machine.
  - cc-eal: Common Criteria EAL2 Provisioning Packages
    (https://ubuntu.com/cc-eal)
- - cis-audit: Center for Internet Security Audit Tools
-   (https://ubuntu.com/cis-audit)
- - esm-apps: UA Apps: Extended Security Maintenance (https://ubuntu.com/esm)
- - esm-infra: UA Infra: Extended Security Maintenance (https://ubuntu.com/esm)
- - fips: NIST-certified FIPS modules (https://ubuntu.com/fips)
+ - cis: Center for Internet Security Audit Tools
+   (https://ubuntu.com/security/certifications#cis)
+ - esm-apps: UA Apps: Extended Security Maintenance (ESM)
+   (https://ubuntu.com/security/esm)
+ - esm-infra: UA Infra: Extended Security Maintenance (ESM)
+   (https://ubuntu.com/security/esm)
  - fips-updates: Uncertified security updates to FIPS modules
- - livepatch: Canonical Livepatch service (https://ubuntu.com/livepatch)
+   (https://ubuntu.com/security/certifications#fips)
+ - fips: NIST-certified FIPS modules
+   (https://ubuntu.com/security/certifications#fips)
+ - livepatch: Canonical Livepatch service
+   (https://ubuntu.com/security/livepatch)
+"""
+)
+
+SERVICES_WRAPPED_HELP = textwrap.dedent(
+    """
+Client to manage Ubuntu Advantage services on a machine.
+ - esm-infra: UA Infra: Extended Security Maintenance (ESM)
+   (https://ubuntu.com/security/esm)
+ - fips-updates: Uncertified security updates to FIPS modules
+   (https://ubuntu.com/security/certifications#fips)
+ - fips: NIST-certified FIPS modules
+   (https://ubuntu.com/security/certifications#fips)
+ - livepatch: Canonical Livepatch service
+   (https://ubuntu.com/security/livepatch)
 """
 )
 
 
-@pytest.fixture(params=["direct", "--help", "ua help"])
+@pytest.fixture(params=["direct", "--help", "ua help", "ua help --all"])
 def get_help(request, capsys):
     if request.param == "direct":
 
@@ -57,7 +81,7 @@ def get_help(request, capsys):
             parser = get_parser()
             help_file = io.StringIO()
             parser.print_help(file=help_file)
-            return help_file.getvalue()
+            return (help_file.getvalue(), "base")
 
     elif request.param == "--help":
 
@@ -67,15 +91,19 @@ def get_help(request, capsys):
                 with pytest.raises(SystemExit):
                     parser.parse_args()
             out, _err = capsys.readouterr()
-            return out
+            return (out, "base")
 
-    elif request.param == "ua help":
+    elif "help" in request.param:
 
         def _get_help_output():
-            with mock.patch("sys.argv", ["ua", "help"]):
+            with mock.patch("sys.argv", request.param.split(" ")):
                 main()
             out, _err = capsys.readouterr()
-            return out
+
+            if "--all" in request.param:
+                return (out, "all")
+
+            return (out, "base")
 
     else:
         raise NotImplementedError("Unknown help source: {}", request.param)
@@ -92,7 +120,7 @@ class TestCLIParser:
         """Help lines are wrapped at 80 chars"""
 
         def cls_mock_factory(desc, url):
-            return mock.Mock(description=desc, help_doc_url=url)
+            return mock.Mock(description=desc, help_doc_url=url, is_beta=False)
 
         m_entitlements.ENTITLEMENT_CLASS_BY_NAME = {
             "test": cls_mock_factory(BIG_DESC, BIG_URL)
@@ -102,11 +130,221 @@ class TestCLIParser:
             " - test: " + " ".join(["123456789"] * 7),
             "   next line ({url})".format(url=BIG_URL),
         ]
-        assert "\n".join(lines) in get_help()
+        out, _ = get_help()
+        assert "\n".join(lines) in out
 
     def test_help_sourced_dynamically_from_each_entitlement(self, get_help):
         """Help output is sourced from entitlement name and description."""
-        assert SERVICES_WRAPPED_HELP in get_help()
+        out, type_request = get_help()
+
+        if type_request == "base":
+            assert SERVICES_WRAPPED_HELP in out
+        else:
+            assert ALL_SERVICES_WRAPPED_HELP in out
+
+    @pytest.mark.parametrize(
+        "out_format, expected_return",
+        (
+            (
+                "tabular",
+                "\n\n".join(
+                    ["Name:\ntest", "Available:\nyes", "Help:\nTest\n\n"]
+                ),
+            ),
+            ("json", {"name": "test", "available": "yes", "help": "Test"}),
+        ),
+    )
+    @mock.patch("uaclient.contract.get_available_resources")
+    @mock.patch(
+        "uaclient.config.UAConfig.is_attached", new_callable=mock.PropertyMock
+    )
+    def test_help_command_when_unnatached(
+        self, m_attached, m_available_resources, out_format, expected_return
+    ):
+        """Test help command for a valid service in an unnatached ua client."""
+        import uaclient.entitlements as ent
+
+        m_args = mock.MagicMock()
+        m_service_name = mock.PropertyMock(return_value="test")
+        type(m_args).service = m_service_name
+        m_format = mock.PropertyMock(return_value=out_format)
+        type(m_args).format = m_format
+        m_all = mock.PropertyMock(return_value=True)
+        type(m_args).all = m_all
+
+        m_entitlement_cls = mock.MagicMock()
+        m_ent_help_info = mock.PropertyMock(return_value="Test")
+        m_entitlement_obj = m_entitlement_cls.return_value
+        type(m_entitlement_obj).help_info = m_ent_help_info
+
+        m_attached.return_value = False
+
+        m_available_resources.return_value = [
+            {"name": "test", "available": True}
+        ]
+
+        fake_stdout = io.StringIO()
+        with mock.patch.object(
+            ent, "ENTITLEMENT_CLASS_BY_NAME", {"test": m_entitlement_cls}
+        ):
+            with contextlib.redirect_stdout(fake_stdout):
+                action_help(m_args, None)
+
+        if out_format == "tabular":
+            assert expected_return.strip() == fake_stdout.getvalue().strip()
+        else:
+            assert expected_return == json.loads(fake_stdout.getvalue())
+
+        assert 1 == m_service_name.call_count
+        assert 1 == m_ent_help_info.call_count
+        assert 1 == m_available_resources.call_count
+        assert 1 == m_attached.call_count
+        assert 1 == m_format.call_count
+
+    @pytest.mark.parametrize(
+        "ent_status, ent_msg",
+        (
+            (status.ContractStatus.ENTITLED, "yes"),
+            (status.ContractStatus.UNENTITLED, "no"),
+        ),
+    )
+    @pytest.mark.parametrize("is_beta", (True, False))
+    @mock.patch("uaclient.contract.get_available_resources")
+    @mock.patch(
+        "uaclient.config.UAConfig.is_attached", new_callable=mock.PropertyMock
+    )
+    def test_help_command_when_attached(
+        self, m_attached, m_available_resources, ent_status, ent_msg, is_beta
+    ):
+        """Test help command for a valid service in an attached ua client."""
+        import uaclient.entitlements as ent
+
+        m_args = mock.MagicMock()
+        m_service_name = mock.PropertyMock(return_value="test")
+        type(m_args).service = m_service_name
+        m_all = mock.PropertyMock(return_value=True)
+        type(m_args).all = m_all
+
+        m_entitlement_cls = mock.MagicMock()
+        m_ent_help_info = mock.PropertyMock(
+            return_value="Test service\nService is being tested"
+        )
+        m_is_beta = mock.PropertyMock(return_value=is_beta)
+        type(m_entitlement_cls).is_beta = m_is_beta
+        m_entitlement_obj = m_entitlement_cls.return_value
+        type(m_entitlement_obj).help_info = m_ent_help_info
+
+        m_entitlement_obj.contract_status.return_value = ent_status
+        m_entitlement_obj.user_facing_status.return_value = (
+            status.UserFacingStatus.ACTIVE,
+            "active",
+        )
+        m_ent_name = mock.PropertyMock(return_value="test")
+        type(m_entitlement_obj).name = m_ent_name
+        m_ent_desc = mock.PropertyMock(return_value="description")
+        type(m_entitlement_obj).description = m_ent_desc
+
+        m_attached.return_value = True
+        m_available_resources.return_value = [
+            {"name": "test", "available": True}
+        ]
+
+        status_msg = "enabled" if ent_msg == "yes" else "—"
+        ufs_call_count = 1 if ent_msg == "yes" else 0
+        ent_name_call_count = 2 if ent_msg == "yes" else 1
+        is_beta_call_count = 1 if status_msg == "enabled" else 0
+
+        expected_msgs = [
+            "Name:\ntest",
+            "Entitled:\n{}".format(ent_msg),
+            "Status:\n{}".format(status_msg),
+        ]
+
+        if is_beta and status_msg == "enabled":
+            expected_msgs.append("Beta:\nTrue")
+
+        expected_msgs.append(
+            "Help:\nTest service\nService is being tested\n\n"
+        )
+
+        expected_msg = "\n\n".join(expected_msgs)
+
+        fake_stdout = io.StringIO()
+        with mock.patch.object(
+            ent, "ENTITLEMENT_CLASS_BY_NAME", {"test": m_entitlement_cls}
+        ):
+            with contextlib.redirect_stdout(fake_stdout):
+                action_help(m_args, None)
+
+        assert expected_msg.strip() == fake_stdout.getvalue().strip()
+        assert 1 == m_service_name.call_count
+        assert 1 == m_ent_help_info.call_count
+        assert 1 == m_available_resources.call_count
+        assert 1 == m_attached.call_count
+        assert 1 == m_ent_desc.call_count
+        assert is_beta_call_count == m_is_beta.call_count
+        assert ent_name_call_count == m_ent_name.call_count
+        assert 1 == m_entitlement_obj.contract_status.call_count
+        assert (
+            ufs_call_count == m_entitlement_obj.user_facing_status.call_count
+        )
+
+    @mock.patch("uaclient.contract.get_available_resources")
+    def test_help_command_for_invalid_service(self, m_available_resources):
+        """Test help command when an invalid service is provided."""
+        m_args = mock.MagicMock()
+        m_service_name = mock.PropertyMock(return_value="test")
+        type(m_args).service = m_service_name
+        m_all = mock.PropertyMock(return_value=True)
+        type(m_args).all = m_all
+
+        m_available_resources.return_value = [
+            {"name": "ent1", "available": True}
+        ]
+
+        fake_stdout = io.StringIO()
+        with contextlib.redirect_stdout(fake_stdout):
+            with pytest.raises(UserFacingError) as excinfo:
+                action_help(m_args, None)
+
+        assert "No help available for 'test'" == str(excinfo.value)
+        assert 1 == m_service_name.call_count
+        assert 1 == m_available_resources.call_count
+
+
+M_PATH_UACONFIG = "uaclient.config.UAConfig."
+
+
+class TestAssertLockFile:
+    @mock.patch("os.getpid", return_value=123)
+    @mock.patch(M_PATH_UACONFIG + "delete_cache_key")
+    @mock.patch(M_PATH_UACONFIG + "add_notice")
+    @mock.patch(M_PATH_UACONFIG + "write_cache")
+    def test_assert_root_creates_lock_and_notice(
+        self,
+        m_write_cache,
+        m_add_notice,
+        m_remove_notice,
+        _m_getpid,
+        FakeConfig,
+    ):
+        arg, kwarg = mock.sentinel.arg, mock.sentinel.kwarg
+
+        @assert_lock_file("some operation")
+        def test_function(args, cfg):
+            assert arg == mock.sentinel.arg
+            assert kwarg == mock.sentinel.kwarg
+
+            return mock.sentinel.success
+
+        ret = test_function(arg, FakeConfig())
+        assert mock.sentinel.success == ret
+        lock_msg = "Operation in progress: some operation"
+        assert [mock.call("", lock_msg)] == m_add_notice.call_args_list
+        assert [mock.call("lock")] == m_remove_notice.call_args_list
+        assert [
+            mock.call("lock", "123:some operation")
+        ] == m_write_cache.call_args_list
 
 
 class TestAssertRoot:
@@ -194,45 +432,20 @@ class TestAssertNotAttached:
         assert "" == out.strip()
 
 
-class TestRequireValidEntitlementName:
-    # rven used in test names as short hand for require_valid_entitlement_name
-    def test_rven_without_name(self):
-        @require_valid_entitlement_name("operation")
-        def test_function(args, cfg):
-            return mock.sentinel.success
-
-        assert mock.sentinel.success == test_function(object(), object())
-
-    def test_rven_with_valid_name(self):
-        @require_valid_entitlement_name("operation")
-        def test_function(args, cfg):
-            return mock.sentinel.success
-
-        m_args = mock.Mock()
-        m_args.name = "esm-infra"
-        assert mock.sentinel.success == test_function(m_args, object())
-
-    @pytest.mark.parametrize("operation_name", ["operation1", "operation2"])
-    def test_rven_with_invalid_name(self, operation_name):
-        @require_valid_entitlement_name(operation_name)
-        def test_function(args, cfg):
-            pass
-
-        m_args = mock.Mock()
-        m_args.name = "invalid_entitlement"
-        with pytest.raises(UserFacingError) as excinfo:
-            test_function(m_args, object())
-
-        assert (
-            "Cannot {} 'invalid_entitlement'".format(operation_name)
-            in excinfo.value.msg
-        )
-
-
 class TestMain:
     @pytest.mark.parametrize(
         "exception,expected_error_msg,expected_log",
         (
+            (
+                LockHeldError(
+                    pid="123",
+                    lock_request="ua reboot-cmds",
+                    lock_holder="ua auto-attach",
+                ),
+                "Unable to perform: ua reboot-cmds.\nOperation in progress:"
+                " ua auto-attach (pid:123)\n",
+                "LockHeldError",
+            ),
             (
                 KeyboardInterrupt,
                 "Interrupt received; exiting.\n",
@@ -245,12 +458,14 @@ class TestMain:
             ),
         ),
     )
+    @mock.patch(M_PATH_UACONFIG + "delete_cache_key")
     @mock.patch("uaclient.cli.setup_logging")
     @mock.patch("uaclient.cli.get_parser")
     def test_errors_handled_gracefully(
         self,
         m_get_parser,
         _m_setup_logging,
+        m_delete_cache_key,
         capsys,
         logging_sandbox,
         caplog_text,
@@ -264,6 +479,7 @@ class TestMain:
         with pytest.raises(SystemExit) as excinfo:
             with mock.patch("sys.argv", ["/usr/bin/ua", "subcmd"]):
                 main()
+        assert 0 == m_delete_cache_key.call_count
 
         exc = excinfo.value
         assert 1 == exc.code
@@ -379,6 +595,23 @@ class TestMain:
 
         assert "['some', 'args']" in log
 
+    def test_argparse_errors_well_formatted(self, capsys):
+        parser = get_parser()
+        with mock.patch("sys.argv", ["ua", "enable"]):
+            with pytest.raises(SystemExit) as excinfo:
+                parser.parse_args()
+        assert 2 == excinfo.value.code
+        _, err = capsys.readouterr()
+        assert (
+            textwrap.dedent(
+                """\
+            usage: ua enable <service> [<service>] [flags]
+            the following arguments are required: service
+        """
+            )
+            == str(err)
+        )
+
 
 class TestSetupLogging:
     @pytest.mark.parametrize("level", (logging.INFO, logging.ERROR))
@@ -458,3 +691,24 @@ class TestSetupLogging:
         assert "after setup" in log_content
         if pre_existing:
             assert "existing content" in log_content
+
+
+class TestGetValidEntitlementNames:
+    @mock.patch("uaclient.cli.entitlements")
+    def test_get_valid_entitlements(self, m_entitlements):
+        m_entitlements.ENTITLEMENT_CLASS_BY_NAME = {
+            "ent1": True,
+            "ent2": True,
+            "ent3": True,
+        }
+
+        service = ["ent1", "ent3", "ent4"]
+        expected_ents_found = ["ent1", "ent3"]
+        expected_ents_not_found = ["ent4"]
+
+        actual_ents_found, actual_ents_not_found = get_valid_entitlement_names(
+            service
+        )
+
+        assert expected_ents_found == actual_ents_found
+        assert expected_ents_not_found == actual_ents_not_found
