@@ -15,17 +15,30 @@ from uaclient.config import (
     PRIVATE_SUBDIR,
     UAConfig,
     parse_config,
+    depth_first_merge_overlay_dict,
 )
 from uaclient.entitlements import (
     ENTITLEMENT_CLASSES,
     ENTITLEMENT_CLASS_BY_NAME,
 )
-from uaclient.status import ContractStatus, UserFacingStatus
+from uaclient.status import (
+    ContractStatus,
+    UserFacingStatus,
+    UserFacingConfigStatus,
+    MESSAGE_ENABLE_REBOOT_REQUIRED_TMPL,
+)
 
 
-KNOWN_DATA_PATHS = (("machine-token", "machine-token.json"),)
+KNOWN_DATA_PATHS = (
+    ("machine-access-cis", "machine-access-cis.json"),
+    ("machine-token", "machine-token.json"),
+)
 M_PATH = "uaclient.entitlements."
 
+DEFAULT_CFG_STATUS = {
+    "configStatus": DEFAULT_STATUS["configStatus"],
+    "configStatusDetails": DEFAULT_STATUS["configStatusDetails"],
+}
 
 ALL_RESOURCES_AVAILABLE = [
     {"name": name, "available": True} for name in ENTITLEMENT_CLASS_BY_NAME
@@ -42,6 +55,49 @@ RESP_ONLY_FIPS_RESOURCE_AVAILABLE = [
 ]
 
 
+class TestNotices:
+    @pytest.mark.parametrize(
+        "notices,expected",
+        (
+            ([], []),
+            ([["a", "a1"]], [["a", "a1"]]),
+            ([["a", "a1"], ["a", "a1"]], [["a", "a1"]]),
+        ),
+    )
+    def test_add_notice_avoids_duplicates(self, notices, expected, tmpdir):
+        cfg = UAConfig({"data_dir": tmpdir.strpath})
+        assert None is cfg.read_cache("notices")
+        for notice in notices:
+            cfg.add_notice(*notice)
+        if notices:
+            assert expected == cfg.read_cache("notices")
+        else:
+            assert None is cfg.read_cache("notices")
+
+    @pytest.mark.parametrize(
+        "notices,removes,expected",
+        (
+            ([], [["a", "a1"]], None),
+            ([["a", "a1"]], [["a", "a1"]], None),
+            ([["a", "a1"], ["a", "a2"]], [["a", "a1"]], [["a", "a2"]]),
+            (
+                [["a", "a1"], ["a", "a2"], ["b", "b2"]],
+                [["a", ".*"]],
+                [["b", "b2"]],
+            ),
+        ),
+    )
+    def test_remove_notice_removes_matching(
+        self, notices, removes, expected, tmpdir
+    ):
+        cfg = UAConfig({"data_dir": tmpdir.strpath})
+        for notice in notices:
+            cfg.add_notice(*notice)
+        for label, descr in removes:
+            cfg.remove_notice(label, descr)
+        assert expected == cfg.read_cache("notices")
+
+
 class TestEntitlements:
     def test_entitlements_property_keyed_by_entitlement_name(self, tmpdir):
         """Return machine_token resourceEntitlements, keyed by name."""
@@ -161,6 +217,30 @@ class TestDataPath:
 
 
 class TestWriteCache:
+    @pytest.mark.parametrize(
+        "key,clears_cache",
+        (
+            ("machine-token", True),
+            ("machine-access-cis", True),
+            ("lock", False),
+        ),
+    )
+    def test_write_cache_clears_machine_token_and_entitlements_instance_vars(
+        self, key, clears_cache, tmpdir
+    ):
+        """Clear _machine_token and _entitlements when machine keys updated."""
+        cfg = UAConfig({"data_dir": tmpdir.strpath})
+        # setup cached values
+        cfg._machine_token = mock.sentinel.token
+        cfg._entitlements = mock.sentinel.entitlements
+        cfg.write_cache(key, "something")
+        if clears_cache:
+            assert None is cfg._entitlements
+            assert None is cfg._machine_token
+        else:
+            assert mock.sentinel.token is cfg._machine_token
+            assert mock.sentinel.entitlements is cfg._entitlements
+
     @pytest.mark.parametrize(
         "key,content",
         (("unknownkey", "content1"), ("another-one", "content2")),
@@ -288,6 +368,63 @@ class TestReadCache:
         assert {"dt": datetime.datetime(2019, 7, 25, 14, 35, 51)} == actual
 
 
+class TestDeleteCacheKey:
+    @pytest.mark.parametrize("property_name", ("status-cache", "lock"))
+    def test_delete_cache_key_removes_public_or_private_data_path_files(
+        self, property_name, tmpdir
+    ):
+        cfg = UAConfig({"data_dir": tmpdir.strpath})
+        cfg.write_cache(property_name, "himom")
+        assert True is os.path.exists(cfg.data_path(property_name))
+        cfg.delete_cache_key(property_name)
+        assert False is os.path.exists(cfg.data_path(property_name))
+        assert None is cfg.read_cache(property_name)
+
+    @pytest.mark.parametrize(
+        "property_name,clears_cache",
+        (
+            ("machine-token", True),
+            ("machine-access-cis", True),
+            ("machine", False),
+        ),
+    )
+    def test_delete_cache_key_clears_machine_token_and_entitlements(
+        self, property_name, clears_cache, tmpdir
+    ):
+        cfg = UAConfig({"data_dir": tmpdir.strpath})
+        token = {
+            "availableResources": ALL_RESOURCES_AVAILABLE,
+            "machineTokenInfo": {
+                "contractInfo": {
+                    "resourceEntitlements": [
+                        {"type": "entitlement1", "entitled": True},
+                        {"type": "entitlement2", "entitled": True},
+                    ]
+                }
+            },
+        }
+        cfg.write_cache("machine-token", token)
+        cfg.entitlements  # sets config _entitlements and _machine_token cache
+        assert cfg._entitlements is not None
+        assert cfg._machine_token is not None
+        cfg.delete_cache_key(property_name)
+        if clears_cache:
+            # internal cache is cleared
+            assert cfg._entitlements is None
+            assert cfg._machine_token is None
+
+        # Reconstitutes _entitlements and _machine_token caches
+        entitlements = cfg.entitlements
+        if property_name == "machine-token":
+            # We performed delete_cache_key("machine-token") above, so None now
+            assert None is cfg._entitlements
+            assert None is cfg.machine_token
+        else:
+            # re-constitute from cache
+            assert entitlements is cfg._entitlements
+            assert cfg._machine_token is cfg.machine_token
+
+
 class TestDeleteCache:
     @pytest.mark.parametrize(
         "property_name,data_path_name,expected_null_value",
@@ -345,9 +482,14 @@ class TestDeleteCache:
         """Any cached files defined in cfg.data_paths will be removed."""
         cfg = UAConfig({"data_dir": tmpdir.strpath})
         # Create half of the cached files, but not all
-        odd_keys = list(cfg.data_paths.keys())[::2]
+        odd_keys = list(sorted(cfg.data_paths.keys()))[::2]
         for odd_key in odd_keys:
-            cfg.write_cache(odd_key, odd_key)
+            if odd_key == "notices":
+                # notices key expects specific list or lists format
+                value = [[odd_key, odd_key]]
+            else:
+                value = odd_key
+            cfg.write_cache(odd_key, value)
 
         present_files = list(
             itertools.chain(
@@ -362,7 +504,7 @@ class TestDeleteCache:
             )
         )
         assert 0 == len(dirty_files), "{} files not deleted".format(
-            len(dirty_files)
+            ", ".join(dirty_files)
         )
 
     def test_delete_cache_ignores_files_not_defined_in_data_paths(
@@ -384,26 +526,81 @@ class TestDeleteCache:
 
 
 class TestStatus:
+    esm_desc = ENTITLEMENT_CLASS_BY_NAME["esm-infra"].description
+    cc_eal_desc = ENTITLEMENT_CLASS_BY_NAME["cc-eal"].description
+
+    def check_beta(self, cls, show_beta, uacfg=None):
+        if not show_beta:
+            if uacfg:
+                allow_beta = uacfg.cfg.get("features", {}).get(
+                    "allow_beta", False
+                )
+
+                if allow_beta:
+                    return False
+
+            return cls.is_beta
+
+        return False
+
+    @pytest.mark.parametrize(
+        "show_beta,expected_services",
+        (
+            (
+                True,
+                [
+                    {
+                        "available": "no",
+                        "name": "cc-eal",
+                        "description": cc_eal_desc,
+                    },
+                    {
+                        "available": "yes",
+                        "name": "esm-infra",
+                        "description": esm_desc,
+                    },
+                ],
+            ),
+            (
+                False,
+                [
+                    {
+                        "available": "yes",
+                        "name": "esm-infra",
+                        "description": esm_desc,
+                    }
+                ],
+            ),
+        ),
+    )
     @mock.patch("uaclient.contract.get_available_resources")
     @mock.patch("uaclient.config.os.getuid", return_value=0)
     def test_root_unattached(
-        self, _m_getuid, m_get_available_resources, FakeConfig
+        self,
+        _m_getuid,
+        m_get_available_resources,
+        show_beta,
+        expected_services,
+        FakeConfig,
     ):
         """Test we get the correct status dict when unattached"""
         cfg = FakeConfig()
         m_get_available_resources.return_value = [
             {"name": "esm-infra", "available": True},
-            {"name": "fips", "available": False},
+            {"name": "cc-eal", "available": False},
         ]
-        esm_desc = ENTITLEMENT_CLASS_BY_NAME["esm-infra"].description
-        fips_desc = ENTITLEMENT_CLASS_BY_NAME["fips"].description
         expected = copy.deepcopy(DEFAULT_STATUS)
-        expected["services"] = [
-            {"available": "yes", "name": "esm-infra", "description": esm_desc},
-            {"available": "no", "name": "fips", "description": fips_desc},
-        ]
-        assert expected == cfg.status()
+        expected["services"] = expected_services
+        with mock.patch(
+            "uaclient.config.UAConfig._get_config_status"
+        ) as m_get_cfg_status:
+            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
+            assert expected == cfg.status(show_beta=show_beta)
 
+    @pytest.mark.parametrize("show_beta", (True, False))
+    @pytest.mark.parametrize(
+        "features_override", ((None), ({"allow_beta": True}))
+    )
     @pytest.mark.parametrize(
         "avail_res,entitled_res,uf_entitled,uf_status",
         (
@@ -443,6 +640,8 @@ class TestStatus:
         entitled_res,
         uf_entitled,
         uf_status,
+        features_override,
+        show_beta,
         FakeConfig,
     ):
         """Test we get the correct status dict when attached with basic conf"""
@@ -460,6 +659,7 @@ class TestStatus:
                 },
             },
         }
+
         available_resource_response = [
             {
                 "name": cls.name,
@@ -475,6 +675,9 @@ class TestStatus:
             m_get_avail_resources.return_value = available_resource_response
 
         cfg = FakeConfig.for_attached_machine(machine_token=token)
+        if features_override:
+            cfg.override_features(features_override)
+
         expected_services = [
             {
                 "description": cls.description,
@@ -489,6 +692,7 @@ class TestStatus:
                 "description_override": None,
             }
             for cls in entitlements.ENTITLEMENT_CLASSES
+            if not self.check_beta(cls, show_beta, cfg)
         ]
         expected = copy.deepcopy(DEFAULT_STATUS)
         expected.update(
@@ -501,13 +705,21 @@ class TestStatus:
                 "subscription-id": "cid",
             }
         )
-        assert expected == cfg.status()
+        with mock.patch(
+            "uaclient.config.UAConfig._get_config_status"
+        ) as m_get_cfg_status:
+            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
+            assert expected == cfg.status(show_beta=show_beta)
         if avail_res:
             assert m_get_avail_resources.call_count == 0
         else:
             assert m_get_avail_resources.call_count == 1
         # cfg.status() idempotent
-        assert expected == cfg.status()
+        with mock.patch(
+            "uaclient.config.UAConfig._get_config_status"
+        ) as m_get_cfg_status:
+            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
+            assert expected == cfg.status(show_beta=show_beta)
 
     @mock.patch("uaclient.contract.get_available_resources")
     @mock.patch("uaclient.config.os.getuid")
@@ -565,6 +777,10 @@ class TestStatus:
             os.lstat(cfg.data_path("status-cache")).st_mode
         )
 
+    @pytest.mark.parametrize("show_beta", (True, False))
+    @pytest.mark.parametrize(
+        "features_override", ((None), ({"allow_beta": False}))
+    )
     @pytest.mark.parametrize(
         "entitlements",
         (
@@ -591,6 +807,8 @@ class TestStatus:
         m_livepatch_uf_status,
         _m_getuid,
         entitlements,
+        features_override,
+        show_beta,
         FakeConfig,
     ):
         """When attached, return contract and service user-facing status."""
@@ -620,6 +838,8 @@ class TestStatus:
         cfg = FakeConfig.for_attached_machine(
             account_name="accountname", machine_token=token
         )
+        if features_override:
+            cfg.override_features(features_override)
         if not entitlements:
             support_level = status.UserFacingStatus.INAPPLICABLE.value
         else:
@@ -636,6 +856,9 @@ class TestStatus:
             }
         )
         for cls in ENTITLEMENT_CLASSES:
+            if self.check_beta(cls, show_beta, cfg):
+                continue
+
             if cls.name == "livepatch":
                 expected_status = status.UserFacingStatus.ACTIVE.value
                 details = "livepatch details"
@@ -652,7 +875,11 @@ class TestStatus:
                     "description_override": None,
                 }
             )
-        assert expected == cfg.status()
+        with mock.patch(
+            "uaclient.config.UAConfig._get_config_status"
+        ) as m_get_cfg_status:
+            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
+            assert expected == cfg.status(show_beta=show_beta)
         assert len(ENTITLEMENT_CLASSES) - 1 == m_repo_uf_status.call_count
         assert 1 == m_livepatch_uf_status.call_count
 
@@ -688,13 +915,27 @@ class TestStatus:
         assert expected_dt == cfg.status()["expires"]
 
     @mock.patch("uaclient.config.os.getuid")
-    def test_nonroot_user_uses_cache_if_available(self, m_getuid, tmpdir):
+    def test_nonroot_user_uses_cache_and_updates_if_available(
+        self, m_getuid, tmpdir
+    ):
         m_getuid.return_value = 1000
 
         status = {"pass": True}
         cfg = UAConfig({"data_dir": tmpdir.strpath})
+        cfg.write_cache("marker-reboot-cmds", "")  # To indicate a reboot reqd
         cfg.write_cache("status-cache", status)
 
+        # Even non-root users can update configStatus details
+        details = MESSAGE_ENABLE_REBOOT_REQUIRED_TMPL.format(
+            operation="configuration changes"
+        )
+        status.update(
+            {
+                "configStatus": UserFacingConfigStatus.REBOOTREQUIRED.value,
+                "configStatusDetails": details,
+                "notices": [],
+            }
+        )
         assert status == cfg.status()
 
 
@@ -796,3 +1037,205 @@ class TestParseConfig:
                 parse_config()
         expected_msg = "Invalid url in config. contract_url: htp://contract"
         assert expected_msg == excinfo.value.msg
+
+
+class TestFeatures:
+    @pytest.mark.parametrize(
+        "cfg_features,expected",
+        (
+            ({}, {}),
+            (None, {}),
+            ({"feature1": "value1"}, {"feature1": "value1"}),
+            (
+                {"feature1": "value1", "feature2": False},
+                {"feature1": "value1", "feature2": False},
+            ),
+        ),
+    )
+    def test_features_are_a_property_of_uaconfig(self, cfg_features, expected):
+        if cfg_features is None:
+            user_cfg = None
+        else:
+            user_cfg = {"features": cfg_features}
+        cfg = UAConfig(cfg=user_cfg)
+        assert expected == cfg.features
+
+
+class TestMachineTokenOverlay:
+    machine_token_dict = {
+        "availableResources": [
+            {"available": False, "name": "cc-eal"},
+            {"available": True, "name": "esm-infra"},
+            {"available": False, "name": "fips"},
+        ],
+        "machineTokenInfo": {
+            "contractInfo": {
+                "resourceEntitlements": [
+                    {
+                        "type": "cc-eal",
+                        "entitled": False,
+                        "affordances": {
+                            "architectures": [
+                                "amd64",
+                                "ppc64el",
+                                "ppc64le",
+                                "s390x",
+                                "x86_64",
+                            ],
+                            "series": ["xenial"],
+                        },
+                        "directives": {
+                            "additionalPackages": ["ubuntu-commoncriteria"],
+                            "aptKey": "key",
+                            "aptURL": "https://esm.ubuntu.com/cc",
+                            "suites": ["xenial"],
+                        },
+                    },
+                    {
+                        "type": "livepatch",
+                        "entitled": True,
+                        "affordances": {
+                            "architectures": ["amd64", "x86_64"],
+                            "tier": "stable",
+                        },
+                        "directives": {
+                            "caCerts": "",
+                            "remoteServer": "https://livepatch.canonical.com",
+                        },
+                        "obligations": {"enableByDefault": True},
+                    },
+                ]
+            }
+        },
+    }
+
+    @mock.patch("uaclient.util.load_file")
+    @mock.patch("uaclient.config.UAConfig.read_cache")
+    @mock.patch("uaclient.config.os.path.exists", return_value=True)
+    def test_machine_token_update_with_overlay(
+        self, m_path, m_read_cache, m_load_file
+    ):
+        user_cfg = {
+            "features": {"machine_token_overlay": "machine-token-path"}
+        }
+        m_read_cache.return_value = self.machine_token_dict
+
+        remote_server_overlay = "overlay"
+        json_str = json.dumps(
+            {
+                "availableResources": [
+                    {"available": False, "name": "esm-infra"},
+                    {"available": True, "name": "test-overlay"},
+                ],
+                "machineTokenInfo": {
+                    "contractInfo": {
+                        "resourceEntitlements": [
+                            {
+                                "type": "livepatch",
+                                "entitled": False,
+                                "affordances": {"architectures": ["test"]},
+                                "directives": {"remoteServer": "overlay"},
+                            }
+                        ]
+                    }
+                },
+            }
+        )
+        m_load_file.return_value = json_str
+
+        expected = copy.deepcopy(self.machine_token_dict)
+        expected["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
+            1
+        ]["directives"]["remoteServer"] = remote_server_overlay
+        expected["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
+            1
+        ]["affordances"]["architectures"] = ["test"]
+        expected["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
+            1
+        ]["entitled"] = False
+        expected["availableResources"][1]["available"] = False
+        expected["availableResources"].append(
+            {"available": True, "name": "test-overlay"}
+        )
+
+        cfg = UAConfig(cfg=user_cfg)
+        print(expected)
+        print(cfg.machine_token)
+        assert expected == cfg.machine_token
+
+    @mock.patch("uaclient.config.UAConfig.read_cache")
+    def test_machine_token_without_overlay(self, m_read_cache):
+        user_cfg = {}
+        m_read_cache.return_value = self.machine_token_dict
+        cfg = UAConfig(cfg=user_cfg)
+        assert self.machine_token_dict == cfg.machine_token
+
+    @mock.patch("uaclient.config.UAConfig.read_cache")
+    @mock.patch("uaclient.config.os.path.exists", return_value=False)
+    def test_machine_token_overlay_file_not_found(self, m_path, m_read_cache):
+        invalid_path = "machine-token-path"
+        user_cfg = {"features": {"machine_token_overlay": invalid_path}}
+        m_read_cache.return_value = self.machine_token_dict
+
+        cfg = UAConfig(cfg=user_cfg)
+        expected_msg = status.INVALID_PATH_FOR_MACHINE_TOKEN_OVERLAY.format(
+            file_path=invalid_path
+        )
+
+        with pytest.raises(exceptions.UserFacingError) as excinfo:
+            cfg.machine_token
+
+        assert expected_msg == str(excinfo.value)
+
+    @mock.patch("uaclient.util.load_file")
+    @mock.patch("uaclient.config.UAConfig.read_cache")
+    @mock.patch("uaclient.config.os.path.exists", return_value=True)
+    def test_machine_token_overlay_json_decode_error(
+        self, m_path, m_read_cache, m_load_file
+    ):
+        invalid_json_path = "machine-token-path"
+        user_cfg = {"features": {"machine_token_overlay": invalid_json_path}}
+        m_read_cache.return_value = self.machine_token_dict
+
+        json_str = '{"directives": {"remoteServer": "overlay"}'
+        m_load_file.return_value = json_str
+        expected_msg = status.ERROR_JSON_DECODING_IN_FILE.format(
+            error="Expecting ',' delimiter: line 1 column 43 (char 42)",
+            file_path=invalid_json_path,
+        )
+
+        cfg = UAConfig(cfg=user_cfg)
+        with pytest.raises(exceptions.UserFacingError) as excinfo:
+            cfg.machine_token
+
+        assert expected_msg == str(excinfo.value)
+
+
+class TestDepthFirstMergeOverlayDict:
+    @pytest.mark.parametrize(
+        "base_dict, overlay_dict, expected_dict",
+        [
+            ({"a": 1, "b": 2}, {"c": 3}, {"a": 1, "b": 2, "c": 3}),
+            (
+                {"a": 1, "b": {"c": 2, "d": 3}},
+                {"a": 1, "b": {"c": 10}},
+                {"a": 1, "b": {"c": 10, "d": 3}},
+            ),
+            (
+                {"a": 1, "b": {"c": 2, "d": 3}},
+                {"d": {"f": 20}},
+                {"a": 1, "b": {"c": 2, "d": 3}, "d": {"f": 20}},
+            ),
+            ({"a": 1, "b": 2}, {}, {"a": 1, "b": 2}),
+            ({"a": 1, "b": 2}, {"a": "test"}, {"a": "test", "b": 2}),
+            ({}, {"a": 1, "b": 2}, {"a": 1, "b": 2}),
+            ({"a": []}, {"a": [1, 2, 3]}, {"a": [1, 2, 3]}),
+            ({"a": [5, 6]}, {"a": [1, 2, 3]}, {"a": [1, 2, 3]}),
+            ({"a": [{"b": 1}]}, {"a": [{"c": 2}]}, {"a": [{"b": 1, "c": 2}]}),
+        ],
+    )
+    def test_depth_first_merge_dict(
+        self, base_dict, overlay_dict, expected_dict
+    ):
+        depth_first_merge_overlay_dict(base_dict, overlay_dict)
+        assert expected_dict == base_dict
