@@ -1,4 +1,5 @@
 import copy
+import logging
 import mock
 import pytest
 import socket
@@ -18,7 +19,7 @@ from uaclient.contract import (
 from uaclient import exceptions
 from uaclient import util
 from uaclient.status import (
-    MESSAGE_CONTRACT_EXPIRED_ERROR,
+    MESSAGE_ATTACH_EXPIRED_TOKEN,
     MESSAGE_ATTACH_FAILURE_DEFAULT_SERVICES,
     MESSAGE_ATTACH_INVALID_TOKEN,
     MESSAGE_UNEXPECTED_ERROR,
@@ -33,12 +34,12 @@ M_REPO_PATH = "uaclient.entitlements.repo.RepoEntitlement."
 
 
 @mock.patch("uaclient.serviceclient.UAServiceClient.request_url")
+@mock.patch("uaclient.contract.util.get_machine_id")
 class TestUAContractClient:
     @pytest.mark.parametrize(
         "detach,expected_http_method",
         ((None, "POST"), (False, "POST"), (True, "DELETE")),
     )
-    @mock.patch("uaclient.contract.util.get_machine_id")
     @mock.patch("uaclient.contract.util.get_platform_info")
     def test__request_machine_token_update(
         self,
@@ -83,6 +84,29 @@ class TestUAContractClient:
             mock.call("/v1/contracts/cId/context/machines/machineId", **params)
         ] == request_url.call_args_list
 
+    def test_request_resource_machine_access(
+        self, get_machine_id, request_url, FakeConfig
+    ):
+        """GET from resource-machine-access route to "enable" a service"""
+        get_machine_id.return_value = "machineId"
+        request_url.return_value = ("response", {})
+        cfg = FakeConfig.for_attached_machine()
+        client = UAContractClient(cfg)
+        kwargs = {"machine_token": "mToken", "resource": "cis"}
+        assert "response" == client.request_resource_machine_access(**kwargs)
+        assert "response" == cfg.read_cache("machine-access-cis")
+        params = {
+            "headers": {
+                "user-agent": "UA-Client/{}".format(get_version()),
+                "accept": "application/json",
+                "content-type": "application/json",
+                "Authorization": "Bearer mToken",
+            }
+        }
+        assert [
+            mock.call("/v1/resources/cis/context/machines/machineId", **params)
+        ] == request_url.call_args_list
+
 
 class TestProcessEntitlementDeltas:
     def test_error_on_missing_entitlement_type(self):
@@ -223,10 +247,17 @@ class TestRequestUpdatedContract:
         )
         assert expected_msg == str(exc.value)
 
+    @pytest.mark.parametrize(
+        "error_code, error_msg",
+        (
+            (401, MESSAGE_ATTACH_INVALID_TOKEN),
+            (403, MESSAGE_ATTACH_EXPIRED_TOKEN),
+        ),
+    )
     @mock.patch("uaclient.util.get_machine_id", return_value="mid")
     @mock.patch(M_PATH + "UAContractClient")
     def test_invalid_token_user_facing_error_on_invalid_token_refresh_failure(
-        self, client, get_machine_id, FakeConfig
+        self, client, get_machine_id, FakeConfig, error_code, error_msg
     ):
         """When attaching, invalid token errors result in proper user error."""
 
@@ -235,11 +266,12 @@ class TestRequestUpdatedContract:
             fake_client._responses = {
                 API_V1_CONTEXT_MACHINE_TOKEN: ContractAPIError(
                     util.UrlError(
-                        "Server error", code=500, url="http://me", headers={}
+                        "Server error",
+                        code=error_code,
+                        url="http://me",
+                        headers={},
                     ),
-                    error_response={
-                        "message": "invalid token: checksum error"
-                    },
+                    error_response={"message": "unauthorized"},
                 )
             }
             return fake_client
@@ -249,7 +281,7 @@ class TestRequestUpdatedContract:
         with pytest.raises(exceptions.UserFacingError) as exc:
             request_updated_contract(cfg, contract_token="yep")
 
-        assert MESSAGE_ATTACH_INVALID_TOKEN == str(exc.value)
+        assert error_msg == str(exc.value)
 
     @mock.patch("uaclient.util.get_machine_id", return_value="mid")
     @mock.patch(M_PATH + "UAContractClient")
@@ -452,52 +484,53 @@ class TestRequestUpdatedContract:
                     }
                 },
                 allow_enable=False,
+                series_overrides=True,
             ),
             mock.call(
                 {"entitlement": {"entitled": False, "type": "ent2"}},
                 {"entitlement": {"entitled": False, "type": "ent2"}},
                 allow_enable=False,
+                series_overrides=True,
             ),
         ]
         assert process_calls == process_entitlement_delta.call_args_list
 
-    @mock.patch(M_PATH + "process_entitlement_delta")
-    @mock.patch("uaclient.util.get_machine_id", return_value="mid")
-    @mock.patch(M_PATH + "UAContractClient")
-    def test_attached_config_refresh_errors_on_expired_contract(
-        self, client, get_machine_id, process_entitlement_delta, FakeConfig
+
+class TestDetachMachineFromContract:
+    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
+    @pytest.mark.parametrize(
+        "curr_machine_id,past_machine_id",
+        (("123", "124"), (123, "124"), ("123", 124), ("123", "123")),
+    )
+    @mock.patch.object(UAContractClient, "_request_machine_token_update")
+    @mock.patch.object(UAContractClient, "_get_platform_data")
+    def test_do_not_make_make_detach_call_when_machine_id_is_different(
+        self,
+        m_platform_data,
+        m_request_machine_token_update,
+        curr_machine_id,
+        past_machine_id,
+        caplog_text,
+        FakeConfig,
     ):
-        """Error when refreshing contract parses an expired contract token."""
+        m_platform_data.return_value = {"machineId": curr_machine_id}
+        cfg = FakeConfig.for_attached_machine()
+        cfg.write_cache("machine-id", past_machine_id)
+        client = UAContractClient(cfg)
 
-        machine_token = {
-            "machineToken": "mToken",
-            "machineTokenInfo": {
-                "contractInfo": {
-                    "effectiveTo": "2018-07-18T00:00:00Z",  # Expired date
-                    "id": "cid",
-                    "resourceEntitlements": [
-                        {"entitled": False, "type": "ent2"},
-                        {"entitled": True, "type": "ent1"},
-                    ],
-                }
-            },
-        }
-        new_token = copy.deepcopy(machine_token)
-        new_token["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
-            1
-        ]["new"] = "newval"
+        actual_value = client.detach_machine_from_contract(
+            machine_token="machine_token",
+            contract_id="contract_id",
+            machine_id="machine_id",
+        )
 
-        def fake_contract_client(cfg):
-            client = FakeContractClient(cfg)
-            client._responses = {self.refresh_route: new_token}
-            return client
-
-        client.side_effect = fake_contract_client
-        cfg = FakeConfig.for_attached_machine(machine_token=machine_token)
-        with pytest.raises(exceptions.UserFacingError) as exc:
-            request_updated_contract(cfg)
-        assert MESSAGE_CONTRACT_EXPIRED_ERROR == str(exc.value)
-        assert new_token == cfg.read_cache("machine-token")
-
-        # No deltas are processed when contract is expired
-        assert 0 == process_entitlement_delta.call_count
+        expected_msg = """\
+        Found new machine-id. Do not call detach on contract backend
+        """
+        if str(past_machine_id) != str(curr_machine_id):
+            assert actual_value == {}
+            assert m_request_machine_token_update.call_count == 0
+            assert expected_msg.strip() in caplog_text()
+        else:
+            assert m_request_machine_token_update.call_count == 1
+            assert expected_msg.strip() not in caplog_text()
