@@ -9,7 +9,7 @@ import uuid
 import mock
 import pytest
 
-from uaclient import cli, util
+from uaclient import cli, util, exceptions, status
 
 PRIVACY_POLICY_URL = (
     "https://www.ubuntu.com/legal/terms-and-policies/privacy-policy"
@@ -125,6 +125,8 @@ class TestIsContainer:
         """Return True when systemd-detect virt exits success."""
         m_subp.return_value = "", ""
         assert True is util.is_container()
+        # Second call for lru_cache test
+        util.is_container()
         calls = [mock.call(["systemd-detect-virt", "--quiet", "--container"])]
         assert calls == m_subp.call_args_list
 
@@ -220,7 +222,8 @@ class TestSubp:
 
 
 class TestParseOSRelease:
-    def test_parse_os_release(self, tmpdir):
+    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
+    def test_parse_os_release(self, caplog_text, tmpdir):
         """parse_os_release returns a dict of values from /etc/os-release."""
         release_file = tmpdir.join("os-release")
         release_file.write(OS_RELEASE_TRUSTY)
@@ -236,6 +239,19 @@ class TestParseOSRelease:
             "VERSION_ID": "14.04",
         }
         assert expected == util.parse_os_release(release_file.strpath)
+        # Add a 2nd call for lru_cache test
+        util.parse_os_release(release_file.strpath)
+
+        os_release_reads = len(
+            [
+                line
+                for line in caplog_text().splitlines()
+                if "os-release" in line
+            ]
+        )
+        assert (
+            1 == os_release_reads
+        ), "lru_cache expected 1 read but found {}".format(os_release_reads)
 
 
 class TestGetPlatformInfo:
@@ -359,6 +375,12 @@ class TestGetMachineId:
             value = util.get_machine_id(
                 data_dir=tmpdir.join("non-existent").strpath
             )
+            # Test lru_cache caches /etc/machine-id from first read
+            etc_machine_id.write("does-not-change")
+            cached_value = util.get_machine_id(
+                data_dir=tmpdir.join("non-existent").strpath
+            )
+            assert value == cached_value
         assert "etc-machine-id" == value
 
     def test_get_machine_id_from_var_lib_dbus_machine_id(self, tmpdir):
@@ -546,7 +568,63 @@ class TestPromptForConfirmation:
         m_input.return_value = user_input
         assert return_value == util.prompt_for_confirmation()
 
-    def test_prompt_text(self, m_input):
-        util.prompt_for_confirmation()
+    @pytest.mark.parametrize(
+        "assume_yes,message,input_calls",
+        [
+            (True, "message ignored on assume_yes=True", []),
+            (False, "", [mock.call("Are you sure? (y/N) ")]),
+            (False, "Custom yep? (y/N) ", [mock.call("Custom yep? (y/N) ")]),
+        ],
+    )
+    def test_prompt_text(self, m_input, assume_yes, message, input_calls):
+        util.prompt_for_confirmation(msg=message, assume_yes=assume_yes)
 
-        assert [mock.call("Are you sure? (y/N) ")] == m_input.call_args_list
+        assert input_calls == m_input.call_args_list
+
+
+class TestIsConfigValueTrue:
+    @pytest.mark.parametrize(
+        "config_dict, return_val",
+        [
+            ({}, False),
+            ({}, False),
+            ({"allow_beta": "true"}, True),
+            ({"allow_beta": "True"}, True),
+            ({"allow_beta": "false"}, False),
+            ({"allow_beta": "False"}, False),
+        ],
+    )
+    def test_is_config_value_true(self, config_dict, return_val, FakeConfig):
+        cfg = FakeConfig()
+        cfg.override_features(config_dict)
+        actual_val = util.is_config_value_true(
+            config=cfg.cfg, path_to_value="features.allow_beta"
+        )
+        assert return_val == actual_val
+
+    @pytest.mark.parametrize(
+        "config_dict, key_val",
+        [
+            ({"allow_beta": "tru"}, "tru"),
+            ({"allow_beta": "Tre"}, "Tre"),
+            ({"allow_beta": "flse"}, "flse"),
+            ({"allow_beta": "Fale"}, "Fale"),
+        ],
+    )
+    def test_exception_is_config_value_true(
+        self, config_dict, key_val, FakeConfig
+    ):
+        path_to_value = "features.allow_beta"
+        cfg = FakeConfig()
+        cfg.override_features(config_dict)
+        with pytest.raises(exceptions.UserFacingError) as excinfo:
+            util.is_config_value_true(
+                config=cfg.cfg, path_to_value=path_to_value
+            )
+
+        expected_msg = status.ERROR_INVALID_CONFIG_VALUE.format(
+            path_to_value=path_to_value,
+            expected_value="boolean string: true or false",
+            value=key_val,
+        )
+        assert expected_msg == str(excinfo.value)
