@@ -10,9 +10,16 @@ from urllib import error, request
 from urllib.parse import urlparse
 import uuid
 from contextlib import contextmanager
-from functools import wraps
+from functools import lru_cache, wraps
 from http.client import HTTPMessage  # noqa: F401
 
+from uaclient import exceptions
+from uaclient import status
+
+
+REBOOT_FILE_CHECK_PATH = "/var/run/reboot-required"
+
+
 try:
     from typing import (  # noqa: F401
         Any,
@@ -135,7 +142,9 @@ class DatetimeAwareJSONDecoder(json.JSONDecoder):
         return o
 
 
-def apply_series_overrides(orig_access: "Dict[str, Any]") -> None:
+def apply_series_overrides(
+    orig_access: "Dict[str, Any]", series: str = None
+) -> None:
     """Apply series-specific overrides to an entitlement dict.
 
     This function mutates orig_access dict by applying any series-overrides to
@@ -155,7 +164,7 @@ def apply_series_overrides(orig_access: "Dict[str, Any]") -> None:
             'Expected entitlement access dict. Missing "entitlement" key:'
             " {}".format(orig_access)
         )
-    series_name = get_platform_info()["series"]
+    series_name = get_platform_info()["series"] if series is None else series
     orig_entitlement = orig_access.get("entitlement", {})
     overrides = orig_entitlement.pop("series", {}).pop(series_name, {})
     for key, value in overrides.items():
@@ -279,6 +288,7 @@ def get_dict_deltas(
     return deltas
 
 
+@lru_cache(maxsize=None)
 def get_machine_id(data_dir: str) -> str:
     """Get system's unique machine-id or create our own in data_dir."""
     # Generate, cache our own uuid if not present on the system
@@ -339,6 +349,7 @@ def get_platform_info() -> "Dict[str, str]":
     return platform_info
 
 
+@lru_cache(maxsize=None)
 def is_container(run_path: str = "/run") -> bool:
     """Checks to see if this code running in a container of some sort"""
     try:
@@ -376,6 +387,7 @@ def load_file(filename: str, decode: bool = True) -> str:
     return content.decode("utf-8")
 
 
+@lru_cache(maxsize=None)
 def parse_os_release(release_file: "Optional[str]" = None) -> "Dict[str, str]":
     if not release_file:
         release_file = "/etc/os-release"
@@ -387,14 +399,22 @@ def parse_os_release(release_file: "Optional[str]" = None) -> "Dict[str, str]":
     return data
 
 
-def prompt_for_confirmation() -> bool:
+def prompt_for_confirmation(msg: str = "", assume_yes: bool = False) -> bool:
     """
     Display a confirmation prompt, returning a bool indicating the response
 
+    :param msg: String custom prompt text to emit from input call.
+    :param assume_yes: Boolean set True to skip confirmation input and return
+        True.
+
     This function will only prompt a single time, and defaults to "no" (i.e. it
     returns False).
     """
-    value = input("Are you sure? (y/N) ")
+    if assume_yes:
+        return True
+    if not msg:
+        msg = status.PROMPT_YES_NO
+    value = input(msg)
     if value.lower().strip() in ["y", "yes"]:
         return True
     return False
@@ -435,6 +455,7 @@ def _subp(
     rcs: "Optional[List[int]]" = None,
     capture: bool = False,
     timeout: "Optional[float]" = None,
+    env: "Optional[Dict[str, str]]" = None,
 ) -> "Tuple[str, str]":
     """Run a command and return a tuple of decoded stdout, stderr.
 
@@ -444,6 +465,7 @@ def _subp(
     @param capture: Boolean set True to log the command and response.
     @param timeout: Optional float indicating number of seconds to wait for
         subp to return.
+    @param env: Optional dictionary of environment variable to pass to Popen.
 
     @return: Tuple of utf-8 decoded stdout, stderr
     @raises ProcessExecutionError on invalid command or returncode not in rcs.
@@ -453,11 +475,13 @@ def _subp(
     bytes_args = [
         x if isinstance(x, bytes) else x.encode("utf-8") for x in args
     ]
+    if env:
+        env.update(os.environ)
     if rcs is None:
         rcs = [0]
     try:
         proc = subprocess.Popen(
-            bytes_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
+            bytes_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
         )
         (out, err) = proc.communicate(timeout=timeout)
     except OSError:
@@ -493,6 +517,7 @@ def subp(
     capture: bool = False,
     timeout: "Optional[float]" = None,
     retry_sleeps: "Optional[List[float]]" = None,
+    env: "Optional[Dict[str, str]]" = None,
 ) -> "Tuple[str, str]":
     """Run a command and return a tuple of decoded stdout, stderr.
 
@@ -506,6 +531,8 @@ def subp(
         retries. Specifying a list of [0.5, 1] instructs subp to retry twice
         on failure; sleeping half a second before the first retry and 1 second
         before the next retry.
+     @param env: Optional dictionary of environment variables to provide to
+        subp.
 
     @return: Tuple of utf-8 decoded stdout, stderr
     @raises ProcessExecutionError on invalid command or returncode not in rcs.
@@ -515,7 +542,7 @@ def subp(
     retry_sleeps = retry_sleeps.copy() if retry_sleeps is not None else None
     while True:
         try:
-            out, err = _subp(args, rcs, capture, timeout)
+            out, err = _subp(args, rcs, capture, timeout, env=env)
             break
         except ProcessExecutionError as e:
             if capture:
@@ -559,3 +586,53 @@ def write_file(filename: str, content: str, mode: int = 0o644) -> None:
         fh.write(content.encode("utf-8"))
         fh.flush()
     os.chmod(filename, mode)
+
+
+def remove_file(file_path: str) -> None:
+    """Remove a file if it exists, logging a message about removal."""
+    if os.path.exists(file_path):
+        logging.debug("Removing file: %s", file_path)
+        os.unlink(file_path)
+
+
+def is_config_value_true(config: "Dict[str, Any]", path_to_value: str):
+    """Check if value parameter can be translated into a boolean 'True' value.
+
+    @param config: A config dict representing
+                   /etc/ubuntu-advantange/uaclient.conf
+    @param path_to_value: The path from where the value parameter was
+                          extracted.
+    @return: A boolean value indicating if the value paramater corresponds
+             to a 'True' boolean value.
+    @raises exceptions.UserFacingError when the value provide by the
+            path_to_value parameter can not be translated into either
+            a 'False' or 'True' boolean value.
+    """
+    value = config
+    default_value = {}  # type: Any
+    paths = path_to_value.split(".")
+    leaf_value = paths[-1]
+    for key in paths:
+        if key == leaf_value:
+            default_value = "false"
+
+        value = value.get(key, default_value)
+
+    value_str = str(value)
+    if value_str.lower() == "true":
+        return True
+    elif value_str.lower() == "false":
+        return False
+    else:
+        raise exceptions.UserFacingError(
+            status.ERROR_INVALID_CONFIG_VALUE.format(
+                path_to_value=path_to_value,
+                expected_value="boolean string: true or false",
+                value=value_str,
+            )
+        )
+
+
+def should_reboot() -> bool:
+    """Check if the system needs to be rebooted."""
+    return os.path.exists(REBOOT_FILE_CHECK_PATH)
